"""The artifact types."""

from ... import types

TMixins = types.TMixins
TKwargs = types.TKwargs
PropertyArtifacts = types.PropertyArtifacts
OpenApiSimplePropertyArtifacts = types.OpenApiSimplePropertyArtifacts
ExtensionSimplePropertyTypedDict = types.ExtensionSimplePropertyTypedDict
ExtensionSimplePropertyArtifacts = types.ExtensionSimplePropertyArtifacts
SimplePropertyTypedDict = types.SimplePropertyTypedDict
SimplePropertyArtifacts = types.SimplePropertyArtifacts
OpenApiJsonPropertyTypedDict = types.OpenApiJsonPropertyTypedDict
OpenApiJsonPropertyArtifacts = types.OpenApiJsonPropertyArtifacts
ExtensionJsonPropertyTypedDict = types.ExtensionJsonPropertyTypedDict
ExtensionJsonPropertyArtifacts = types.ExtensionJsonPropertyArtifacts
JsonPropertyTypedDict = types.JsonPropertyTypedDict
JsonPropertyArtifacts = types.JsonPropertyArtifacts
RelationshipPropertyArtifacts = types.RelationshipPropertyArtifacts
NotManyToManyRelationshipPropertyArtifacts = (
    types.NotManyToManyRelationshipPropertyArtifacts
)
OneToManyRelationshipPropertyTypedDict = types.OneToManyRelationshipPropertyTypedDict
OneToManyRelationshipPropertyArtifacts = types.OneToManyRelationshipPropertyArtifacts
XToOneRelationshipPropertyTypedDict = types.XToOneRelationshipPropertyTypedDict
XToOneRelationshipPropertyArtifacts = types.XToOneRelationshipPropertyArtifacts
ManyToOneRelationshipPropertyTypedDict = types.ManyToOneRelationshipPropertyTypedDict
ManyToOneRelationshipPropertyArtifacts = types.ManyToOneRelationshipPropertyArtifacts
OneToOneRelationshipPropertyTypedDict = types.OneToOneRelationshipPropertyTypedDict
OneToOneRelationshipPropertyArtifacts = types.OneToOneRelationshipPropertyArtifacts
ManyToManyRelationshipPropertyTypedDict = types.ManyToManyRelationshipPropertyTypedDict
ManyToManyRelationshipPropertyArtifacts = types.ManyToManyRelationshipPropertyArtifacts
TAnyRelationshipPropertyTypedDict = types.TAnyRelationshipPropertyTypedDict
TAnyRelationshipPropertyArtifacts = types.TAnyRelationshipPropertyArtifacts
BackrefPropertyTypedDict = types.BackrefPropertyTypedDict
BackrefSubType = types.BackrefSubType
BackrefPropertyArtifacts = types.BackrefPropertyArtifacts
TAnyPropertyTypedDict = types.TAnyPropertyTypedDict
TAnyPropertyArtifacts = types.TAnyPropertyArtifacts
PropertyValue = types.PropertyValue
PropertiesValue = types.PropertiesValue
ModelTypedDict = types.ModelTypedDict
ModelValue = types.ModelValue
ModelBackrefArtifacts = types.ModelBackrefArtifacts
ModelExPropertiesArtifacts = types.ModelExPropertiesArtifacts
ModelArtifacts = types.ModelArtifacts
ModelsValue = types.ModelsValue
SpecValue = types.SpecValue
ModelsModelArtifacts = types.ModelsModelArtifacts
