"""Apply automatic code formatting to source code."""


def apply(*, source: str) -> str:
    """
    Apply automatic code formatting to source code.

    Args:
        source: The source code.

    Returns:
        The formatted source code.

    """
    return source
