"""Autogenerated SQLAlchemy models based on OpenAlchemy models."""
# pylint: disable=no-member,super-init-not-called,unused-argument
{% if imports %}

{% for import_ in imports %}
import {{ import_ }}
{% endfor %}
{% endif %}

import sqlalchemy
{% if python_minor_version < 8 %}
import typing_extensions
{% endif %}
from sqlalchemy import orm

from open_alchemy import models

Base = models.Base  # type: ignore
{% for model in models %}
{{ model }}
{% endfor %}
